import numpy as np
from pyBKT.util import check_data
from pyBKT.fit import E_step
from pyBKT.fit import M_step
import os

def EM_fit(model, data, tol = None, maxiter = None, parallel = True):

    if tol is None: 
        tol = 1e-3
    if maxiter is None:
        maxiter = 100

    num_subparts = data["data"].shape[0] #mmm the first dimension of data represents each subpart?? interesting.
    num_resources = len(model["learns"])
    log_likelihoods = np.zeros((maxiter, 1))

    for i in range(maxiter):
        result = E_step.run(data, model, 1, int(parallel))
        for j in range(num_resources):
            result['all_trans_softcounts'][j] = result['all_trans_softcounts'][j].transpose()
        for j in range(num_subparts):
            result['all_emission_softcounts'][j] = result['all_emission_softcounts'][j].transpose()

        log_likelihoods[i][0] = result['total_loglike']

        if(i > 1 and abs(log_likelihoods[i][0] - log_likelihoods[i-1][0]) < tol):
            break

        model = M_step.run(model, result['all_trans_softcounts'], result['all_emission_softcounts'], result['all_initial_softcounts'])

    return(model, log_likelihoods[:i+1])



import numpy as np
from pyBKT.fit import E_step
from pyBKT.fit import predict_onestep_states

# correct_emission_predictions is a  num_subparts x T array, where element
# (i,t) is predicted probability that answer to subpart i at time t+1 is correct
def run(model, data, parallel = True):

    num_subparts = data["data"].shape[0]  # mmm the first dimension of data represents each subpart?? interesting.
    num_resources = len(model["learns"])

    result = E_step.run(data, model, 1, int(parallel))
    for j in range(num_resources):
        result['all_trans_softcounts'][j] = result['all_trans_softcounts'][j].transpose()
    for j in range(num_subparts):
        result['all_emission_softcounts'][j] = result['all_emission_softcounts'][j].transpose()

    state_predictions = predict_onestep_states.run(data, model, result['alpha'], int(parallel))
    p = state_predictions.shape
    state_predictions = state_predictions.flatten(order = 'C').reshape(p, order = 'F')
    # multiguess solution, should work
    correct_emission_predictions = np.expand_dims(model["guesses"], axis = 1) @ np.expand_dims(state_predictions[0,:], axis = 0) + np.expand_dims(1-model["slips"], axis = 1) @ np.expand_dims(state_predictions[1,:], axis = 0)
    #correct_emission_predictions = model['guesses'] * np.asarray([state_predictions[0,:]]).T + (1 - model['slips']) * np.asarray([state_predictions[1,:]]).T
    flattened_predictions = np.take_along_axis(correct_emission_predictions, (data['data'] != 0).argmax(axis = 0)[:, None].T, axis = 0)
    return (flattened_predictions.ravel(), state_predictions)




#########################################
# EM_fit.py                             #
# EM_fit                                #
#                                       #
# @author Anirudhan Badrinath           #
# @author Christian Garay               #
# Last edited: 20 March 2020            #
#########################################

import numpy as np
from time import time
from pyBKT.util import check_data
from pyBKT.fit import M_step
from multiprocessing import Pool, cpu_count

gs = globals()

def EM_fit(model, data, tol = 0.005, maxiter = 100, parallel = True):

    check_data.check_data(data)

    num_subparts = data["data"].shape[0] #mmm the first dimension of data represents each subpart?? interesting.
    num_resources = len(model["learns"])

    trans_softcounts = np.zeros((num_resources,2,2))
    emission_softcounts = np.zeros((num_subparts,2,2))
    init_softcounts = np.zeros((2, 1))
    log_likelihoods = np.zeros((maxiter, 1))

    result = {}
    result['all_trans_softcounts'] = trans_softcounts
    result['all_emission_softcounts'] = emission_softcounts
    result['all_initial_softcounts'] = init_softcounts

    for i in range(maxiter):
        result = run(data, model, result['all_trans_softcounts'], result['all_emission_softcounts'], result['all_initial_softcounts'], 1, parallel)
        for j in range(num_resources):
            result['all_trans_softcounts'][j] = result['all_trans_softcounts'][j].transpose()
        for j in range(num_subparts):
            result['all_emission_softcounts'][j] = result['all_emission_softcounts'][j].transpose()

        log_likelihoods[i][0] = result['total_loglike']
        if(i > 1 and abs(log_likelihoods[i][0] - log_likelihoods[i-1][0]) <= tol):
            break

        model = M_step.run(model, result['all_trans_softcounts'], result['all_emission_softcounts'], result['all_initial_softcounts'])

    return(model, log_likelihoods[:i+1])

def run(data, model, trans_softcounts, emission_softcounts, init_softcounts, num_outputs, parallel = True):
    # Processed Parameters
    alldata = data["data"]
    bigT, num_subparts = len(alldata[0]), len(alldata)
    allresources, starts, learns, forgets, guesses, slips, lengths = \
            data["resources"], data["starts"], model["learns"], model["forgets"], model["guesses"], model["slips"], data["lengths"]

    prior, num_sequences, num_resources = model["prior"], len(starts), len(learns)
    normalizeLengths = False

    initial_distn = np.empty((2, ), dtype = 'float')
    initial_distn[0] = 1 - prior
    initial_distn[1] = prior

    As = np.empty((2, 2 * num_resources))
    interleave(As[0], 1 - learns, forgets.copy())
    interleave(As[1], learns.copy(), 1 - forgets)

    Bn = np.empty((2, 2 * num_subparts))
    interleave(Bn[0], 1 - guesses, guesses.copy())
    interleave(Bn[1], slips.copy(), 1 - slips)

    # Outputs
    all_trans_softcounts = np.zeros((2, 2 * num_resources))
    all_emission_softcounts = np.zeros((2, 2 * num_subparts))
    all_initial_softcounts = np.zeros((2, 1))

    alpha_out = np.zeros((2, bigT))

    total_loglike = np.empty((1,1))
    total_loglike.fill(0)

    input = {"As": As, "Bn": Bn, "initial_distn": initial_distn, 'allresources': allresources, \
             'starts': starts, 
             'lengths': lengths, \
             'num_resources': num_resources, 'num_subparts': num_subparts, \
             'alldata': alldata, 'normalizeLengths': normalizeLengths, 'alpha_out': alpha_out}

    num_threads = cpu_count() if parallel else 1
    thread_counts = [None for i in range(num_threads)]
    for thread_num in range(num_threads):
        blocklen = 1 + ((num_sequences - 1) // num_threads)
        sequence_idx_start = int(blocklen * thread_num)
        sequence_idx_end = min(sequence_idx_start+blocklen, num_sequences)
        thread_counts[thread_num] = {'sequence_idx_start': sequence_idx_start, 'sequence_idx_end': sequence_idx_end}
        thread_counts[thread_num].update(input)

    p = Pool(len(thread_counts))
    x = p.map(inner, thread_counts)
    p.close()

    for i in x:
        total_loglike += i[3]
        all_trans_softcounts += i[0] # + all_trans_softcounts
        all_emission_softcounts += i[1] #+ all_emission_softcounts
        all_initial_softcounts += i[2] #+ all_initial_softcounts
        for sequence_start, T, alpha in i[4]:
            alpha_out[:, sequence_start: sequence_start + T] += alpha
    all_trans_softcounts = all_trans_softcounts.flatten(order = 'F')
    all_emission_softcounts = all_emission_softcounts.flatten(order = 'F')
    result = {}
    result["total_loglike"] = total_loglike;
    result["all_trans_softcounts"] = np.reshape(all_trans_softcounts, (num_resources, 2, 2), order = 'C')
    result["all_emission_softcounts"] = np.reshape(all_emission_softcounts, (num_subparts, 2, 2), order = 'C')
    result["all_initial_softcounts"] = all_initial_softcounts
    result["alpha_out"] = alpha_out.flatten(order = 'F').reshape(alpha_out.shape, order = 'C')

    return result

def interleave(m, v1, v2):
    m[0::2], m[1::2] = v1, v2

def inner(x):
    As, Bn, initial_distn, allresources, starts, lengths, num_resources, num_subparts, alldata, normalizeLengths, alpha_out, sequence_idx_start, sequence_idx_end = \
            x['As'], x['Bn'], x['initial_distn'], x['allresources'], x['starts'], x['lengths'], x['num_resources'], x['num_subparts'], x['alldata'], x['normalizeLengths'], \
            x['alpha_out'], x['sequence_idx_start'], x['sequence_idx_end']
    N_R, N_S = 2 * num_resources, 2 * num_subparts
    trans_softcounts_temp = np.zeros((2, N_R))
    emission_softcounts_temp = np.zeros((2, N_S))
    init_softcounts_temp = np.zeros((2, 1))
    loglike = 0

    alphas = [] 
    dot, sum, log = np.dot, np.sum, np.log

    for sequence_index in range(sequence_idx_start, sequence_idx_end):
        sequence_start = starts[sequence_index] - 1
        T = lengths[sequence_index]

        # likelihood calculation
        likelihoods = np.ones((2, T))
        alpha = np.empty((2, T))
        for t in range(min(2, T)):
            for n in range(num_subparts):
                data_temp = alldata[n][sequence_start + t]
                if data_temp:
                    sl = Bn[:, 2 * n + int(data_temp == 2)]
                    likelihoods[:,t] *= np.where(sl == 0, 1, sl)
        
        # setup for alpha, included in loop for efficiency (to keep it as one loop)
        alpha[:,0] = initial_distn * likelihoods[:,0]
        norm = sum(alpha[:,0])
        alpha[:,0] /= norm
        contribution = log(norm) / (T if normalizeLengths else 1)
        loglike += contribution

        # combined with t = 2 for efficiency, otherwise we need another loop
        if T >= 2:
            resources_temp = allresources[sequence_start]
            k = 2 * (resources_temp - 1)
            alpha[:, 1] = dot(As[0:2, k: k + 2], alpha[:, 0]) * \
                      likelihoods[:, 1]
            norm = sum(alpha[:, 1])
            alpha[:, 1] /= norm
            contribution = log(norm) / (T if normalizeLengths else 1)
            loglike += contribution

        for t in range(2, T):
            for n in range(num_subparts):
                data_temp = alldata[n][sequence_start + t]
                if data_temp:
                    sl = Bn[:, 2 * n + int(data_temp == 2)]
                    likelihoods[:,t] *= np.where(sl == 0, 1, sl)
            # general loop for alpha calculations
            resources_temp = allresources[sequence_start + t - 1]
            k = 2 * (resources_temp - 1)
            alpha[:, t] = dot(As[0:2, k: k + 2], alpha[:, t - 1]) * \
                      likelihoods[:, t]
            norm = sum(alpha[:, t])
            alpha[:, t] /= norm
            loglike += log(norm) / (T if normalizeLengths else 1)

        # backward messages and statistic counting
        gamma = np.empty((2, T))
        gamma[:, (T - 1)] = alpha[:, (T - 1)].copy()

        # copy it to begin with for efficiency
        As_temp = As.copy() 
        # only one pass of the previous update, which is now merged into this loop
        f = True
        for t in range(T - 2, -1, -1):
            resources_temp = allresources[sequence_start + t]
            k = 2 * (resources_temp - 1)
            A = As_temp[0: 2, k: k + 2]
            pair = A.copy() # don't want to modify original A 
            pair[0] *= alpha[:, t]
            pair[1] *= alpha[:, t]
            dotted, gamma_t = dot(A, alpha[:, t]), gamma[:, (t + 1)]
            pair[:, 0] = (pair[:, 0] * gamma_t) / dotted
            pair[:, 1] = (pair[:, 1] * gamma_t) / dotted
            np.nan_to_num(pair, copy = False)
            trans_softcounts_temp[0: 2, k: k + 2] += pair
            gamma[:, t] = sum(pair, axis = 0)
            for n in range(num_subparts):
                data_temp = alldata[n][sequence_start+t]
                if data_temp:
                    emission_softcounts_temp[:, (2 * n + int(data_temp == 2))] += gamma[:, t] 
                if f:
                    data_temp_p = alldata[n][sequence_start + (T-1)]
                    if data_temp_p:
                        emission_softcounts_temp[:, (2 * n + int(data_temp_p == 2))] += gamma[:, (T - 1)]
            f = False
        init_softcounts_temp += gamma[:, 0].reshape((2, 1)) 
        alphas.append((sequence_start, T, alpha))
    return [trans_softcounts_temp, emission_softcounts_temp, init_softcounts_temp, loglike, alphas]



#########################################
# data_helper.py                        #
# data_helper                           #
#                                       #
# @author Frederic Wang                 #
# Last edited: 07 April 2021            #
#########################################

import sys
sys.path.append('../')
import os
import pandas as pd
import numpy as np
import io
import requests

def convert_data(url, skill_name, defaults=None, model_type=None, gs_refs=None, resource_refs=None, return_df = False, folds=False):
    if model_type:
        multilearn, multiprior, multipair, multigs = model_type
    else:
        multilearn, multiprior, multipair, multigs = [False] * 4
    pd.set_option('mode.chained_assignment', None)
    df = None

    if not isinstance(skill_name, str):
        skill_name = '|'.join(skill_name)
    
    if isinstance(url, pd.DataFrame):
        df = url
    else:
        # if url is a local file, read it from there
        if os.path.exists(url):
            try:
                # assume comma delimiter
                df = pd.read_csv(url, low_memory=False, encoding="latin")
            except:
                # try tab delimiter if comma delimiter fails
                df = pd.read_csv(url, low_memory=False, encoding="latin", delimiter='\t')
        
        # otherwise, fetch it from web using requests
        elif url[:4] == "http":
            s = requests.get(url).content
            try:
                df = pd.read_csv(s, low_memory=False, encoding="latin")
            except:
                df = pd.read_csv(s, low_memory=False, encoding="latin", delimiter='\t')
            f = open(url.split('/')[-1], 'w+')
            # save csv to local file for quick lookup in the future
            df.to_csv(f)
        else:
            raise ValueError("File path or dataframe input not found")

    # default column names for assistments
    as_default={'order_id': 'order_id',
                 'skill_name': 'skill_name',
                 'correct': 'correct',
                 'user_id': 'user_id',
                 'multilearn': 'template_id',
                 'multiprior': 'correct',
                 'multipair': 'template_id',
                 'multigs': 'template_id',
                 'folds': 'user_id',
                 }

    # default column names for cognitive tutors
    ct_default={'order_id': 'Row',
                'skill_name': 'KC(Default)',
                'correct': 'Correct First Attempt',
                'user_id': 'Anon Student Id',
                'multilearn': 'Problem Name',
                'multiprior': 'Correct First Attempt',
                'multipair': 'Problem Name',
                'multigs': 'Problem Name',
                'folds': 'Anon Student Id',
                                 }

    # integrate custom defaults with default assistments/ct columns if they are still unspecified
    if defaults is None:
        defaults = {}
    elif isinstance(defaults, dict):
        ks = tuple(defaults.items())
        for k, v in ks:
            if v not in df.columns and k in as_default:
                defaults.pop(k)
    else:
        raise ValueError("incorrectly specified defaults")

    if any(x in list(df.columns) for x in as_default.values()):
        for k,v in as_default.items():
            if k not in defaults and as_default[k] in df.columns:
                defaults[k] = as_default[k]
    if any(x in list(df.columns) for x in ct_default.values()):
        for k,v in ct_default.items():
            if k not in defaults and ct_default[k] in df.columns:
                defaults[k] = ct_default[k]

    # sort by the order in which the problems were answered
    if "order_id" in defaults:
        df[defaults["order_id"]] = df[defaults["order_id"]].apply(lambda x: int(x))
        df.sort_values(defaults["order_id"], inplace=True)
    
    if "user_id" not in defaults:
        raise KeyError("user id default column not specified")
    elif defaults["user_id"] not in df.columns:
        raise KeyError("specified user id default column not in data")
        
    if "correct" not in defaults:
        raise KeyError("correct default column not specified")
    elif defaults["correct"] not in df.columns:
        raise KeyError("specified correct default column not in data")
        
    if "skill_name" not in defaults:
        raise KeyError("skill name default column not specified")
    elif defaults["skill_name"] not in df.columns:
        raise KeyError("specified skill name default column not in data")
    
    # make sure all responses of the same user are grouped together with stable sorting
    df.sort_values(defaults["user_id"], kind="mergesort", inplace=True)
    
    if "original" in df.columns:
        df = df[(df["original"]==1)]
        
    df[defaults["skill_name"]] = df[defaults["skill_name"]].apply(str)
    try:
        df[defaults["correct"]] = df[defaults["correct"]].apply(int)
    except:
        raise ValueError("Invalid Data In Specified Corrects Column")
    
    datas = {}
    skill_name = '^(' + skill_name + ')$'
    all_skills = pd.Series(df[defaults["skill_name"]].unique()).dropna()
    all_skills = all_skills[all_skills.str.match(skill_name).astype(bool)]
    if all_skills.empty:
        raise ValueError("no matching skills")
    for skill_ in all_skills:

        if resource_refs is None or skill_ not in resource_refs:
            resource_ref = None
        else:
            resource_ref = resource_refs[skill_]["resource_names"]
        if gs_refs is None or skill_ not in gs_refs:
            gs_ref = None
        else:
            gs_ref = gs_refs[skill_]["gs_names"]

        # filter out based on skill
        df3 = df[df[defaults["skill_name"]] == skill_]
        if df3.empty:
            raise ValueError("Incorrect Skill or Dataset Specified")

        stored_index = df3.index.copy()

        # convert from 0=incorrect,1=correct to 1=incorrect,2=correct
        if set(df3.loc[:,defaults["correct"]].unique()) - set([-1, 0, 1]) != set():
            raise ValueError("correctness must be -1 (no response), 0 (incorrect), or 1 (correct)")
        df3.loc[:,defaults["correct"]]+=1
        
        # array representing correctness of student answers
        data=np.array(df3[defaults["correct"]])
        
        Data={}
    
        # create starts and lengths arrays
        lengths = np.array(df3.groupby(defaults["user_id"])[defaults["user_id"]].count().values, dtype=np.int64)
        starts = np.zeros(len(lengths), dtype=np.int64)
        starts[0] = 1
        for i in range(1, len(lengths)):
            starts[i] = starts[i-1] + lengths[i-1]

        if multipair + multiprior + multilearn > 1:
            raise ValueError("cannot specify more than 1 resource handling")

        # different types of resources handling: multipair, multiprior, multilearn and n/a
        if multipair:
            if "multipair" not in defaults:
                raise KeyError("multipair default column not specified")
            elif defaults["multipair"] not in df3.columns:
                raise KeyError("specified multipair default column not in data")
                
            resources = np.ones(len(data), dtype=np.int64)
            if resource_ref is None:
                new_resource_ref = {}
                new_resource_ref["Default"] = 1 #no pair
            for i in range(len(df3)):
                # for the first entry of a new student, no pair
                if i == 0 or df3[i:i+1][defaults["user_id"]].values != df3[i-1:i][defaults["user_id"]].values:
                    resources[i] = 1
                else:
                    # each pair is keyed via "[item 1] [item 2]"
                    k = (str)(df3[i:i+1][defaults["multipair"]].values)+" "+(str)(df3[i-1:i][defaults["multipair"]].values)
                    if resource_ref is not None and k not in resource_ref:
                        raise ValueError("Pair", k, "not fitted")
                    if k not in new_resource_ref:
                        # form the resource reference as we iterate through the dataframe, mapping each new pair to a number [1, # total pairs]
                        new_resource_ref[k] = len(new_resource_ref)+1
                    resources[i] = new_resource_ref[k]
            if resource_ref is None:
                resource_ref = new_resource_ref
        elif multiprior:
            if "multiprior" not in defaults:
                raise KeyError("multiprior default column not specified")
            elif defaults["multiprior"] not in df3.columns:
                raise KeyError("specified multiprior default column not in data")
                
            resources = np.ones(len(data)+len(starts), dtype=np.int64)
            new_data = np.zeros(len(data)+len(starts), dtype=np.int32)
            # create new resources [2, #total + 1] based on how student initially responds
            all_priors = df3[defaults["multiprior"]].unique()
            if resource_ref is None:
                resource_ref = dict(zip(all_priors,range(2, len(df3[defaults["multiprior"]].unique())+2)))
                resource_ref["Default"] = 1
            else:
                for i in all_priors:
                    if i not in resource_ref:
                        raise ValueError("Prior", i, "not fitted")
            all_resources = np.array(df3[defaults["multiprior"]].apply(lambda x: resource_ref[x]))
            # create phantom timeslices with resource 2 or 3 in front of each new student based on their initial response
            for i in range(len(starts)):
                new_data[i+starts[i]:i+starts[i]+lengths[i]] = data[starts[i]-1:starts[i]+lengths[i]-1]
                resources[i+starts[i]-1] = 1
                resources[i+starts[i]:i+starts[i]+lengths[i]] = all_resources[starts[i]-1:starts[i]+lengths[i]-1]
                starts[i] += i
                lengths[i] += 1
            data = new_data
        elif multilearn:
            if "multilearn" not in defaults:
                raise KeyError("multilearn default column not specified")
            elif defaults["multilearn"] not in df3.columns:
                raise KeyError("specified multilearn default column not in data")
                
            all_learns = df3[defaults["multilearn"]].unique()
            if resource_ref is None:
                # map each new resource found to a number [1, # total]
                resource_ref=dict(zip(all_learns,range(1,len(df[defaults["multilearn"]].unique())+1)))
            else:
                for i in all_learns:
                    if i not in resource_ref:
                        raise ValueError("Learn rate", i, "not fitted")
                
            resources = np.array(df3[defaults["multilearn"]].apply(lambda x: resource_ref[x]))
        else:
            resources=np.array([1]*len(data))


        # multigs handling, make data n-dimensional where n is number of g/s types
        if multigs:
            if "multigs" not in defaults:
                raise KeyError("multigs default column not specified")
            elif defaults["multigs"] not in df3.columns:
                raise KeyError("specified multigs default column not in data")
                
            all_guess = df3[defaults["multigs"]].unique()
            # map each new guess/slip case to a row [0, # total]
            if gs_ref is None:
                gs_ref=dict(zip(all_guess,range(len(df[defaults["multigs"]].unique()))))
            else:
                for i in all_guess:
                    if i not in gs_ref:
                        raise ValueError("Guess rate", i, "not previously fitted")
            data_ref = np.array(df3[defaults["multigs"]].apply(lambda x: gs_ref[x]))
        
            # make data n-dimensional, fill in corresponding row and make other non-row entries 0
            data_temp = np.zeros((len(df3[defaults["multigs"]].unique()), len(df3)))
            for i in range(len(data_temp[0])):
                data_temp[data_ref[i]][i] = data[i]
            Data["data"]=np.asarray(data_temp,dtype='int32')
        else:
            data = [data]
            Data["data"]=np.asarray(data,dtype='int32')

        # for when no resource and/or guess column is selected
        if not multilearn and not multipair and not multiprior:
            resource_ref = {}
            resource_ref["default"]=1
        if not multigs:
            gs_ref = {}
            gs_ref["default"]=1
            
        Data["starts"]=starts
        Data["lengths"]=lengths
        Data["resources"]=resources
        Data["resource_names"]=resource_ref
        Data["gs_names"]=gs_ref
        Data["index"]=stored_index
        if folds:
            Data["folds"] = np.array(df3[defaults["folds"]])

        datas[skill_] = Data

    if return_df:
        return datas, df

    return datas



import numpy as np
from pyBKT.generate import synthetic_data
from pyBKT.generate import random_model, random_model_uni
from pyBKT.fit import EM_fit
from copy import deepcopy
from pyBKT.util import print_dot

#parameters
num_subparts = 4
num_resources = 2
num_fit_initializations = 25
observation_sequence_lengths = np.full(50, 100, dtype=np.int)

#generate synthetic model and data.
#model is really easy.
truemodel = {}

truemodel["As"] = np.zeros((num_resources, 2, 2), dtype=np.float_)
truemodel["As"][0, :, :] = np.transpose([[0.75, 0.25], [0.1, 0.9]])
truemodel["As"][1, :, :] = np.transpose([[0.9, 0.1], [0.1, 0.9]])
truemodel["learns"] = truemodel["As"][:, 1, 0]
truemodel["forgets"] = truemodel["As"][:, 0, 1]

truemodel["pi_0"] = np.array([[0.9], [0.1]]) #TODO: one prior per resource? does this array needs to be col?
truemodel["prior"] = 0.1

truemodel["guesses"] = np.full(num_subparts, 0.05, dtype=np.float_)
truemodel["slips"] = np.full(num_subparts, 0.25, dtype=np.float_)

truemodel["resources"] = np.random.randint(1, high = num_resources+1, size = sum(observation_sequence_lengths))

#data!
print("generating data...")
data = synthetic_data.synthetic_data(truemodel, observation_sequence_lengths)

#fit models, starting with random initializations
print('fitting! each dot is a new EM initialization')

best_likelihood = float("-inf")
for i in range(num_fit_initializations):
    print_dot.print_dot(i, num_fit_initializations)
    fitmodel = random_model.random_model(num_resources, num_subparts)
    (fitmodel, log_likelihoods) = EM_fit.EM_fit(fitmodel, data)
    if (log_likelihoods[-1] > best_likelihood):
        best_likelihood = log_likelihoods[-1]
        best_model = fitmodel

# compare the fit model to the true model
print('')

print('these two should look similar')
print(truemodel['As'])
print('')
print(best_model['As'])

print('')
print('these should look similar too')
print(1-truemodel['guesses'])
print('')
print(1-best_model['guesses'])

print('')
print('these should look similar too')
print(1-truemodel['slips'])
print('')
print(1-best_model['slips'])


import numpy as np
from pyBKT.util import dirrnd

def random_model(num_resources=None, num_subparts=None, trans_prior=None, given_notknow_prior=None, given_know_prior=None, pi_0_prior=None):

    if num_resources is None: num_resources = 1
    if num_subparts is None: num_subparts = 1

    if trans_prior is None:
        trans_prior = np.tile(np.transpose([[20, 4], [1, 20]]), (num_resources, 1)).reshape((num_resources, 2, 2))
    if given_notknow_prior is None:
        given_notknow_prior = np.tile([[5], [0.5]], (1, num_subparts))
    if given_know_prior is None:
        given_know_prior = np.tile([[0.5], [5]], (1, num_subparts))
    if pi_0_prior is None:
        pi_0_prior = np.array([[100], [1]])

    As = dirrnd.dirrnd(trans_prior)
    given_notknow = dirrnd.dirrnd(given_notknow_prior)
    given_know = dirrnd.dirrnd(given_know_prior)
    #emissions = np.dstack((given_notknow.reshape((num_subparts, 2, 1)), given_know.reshape((num_subparts, 2, 1))))
    emissions = np.stack((np.transpose(given_notknow.reshape((2, num_subparts))), np.transpose(given_know.reshape((2, num_subparts)))), axis=1)
    pi_0 = dirrnd.dirrnd(pi_0_prior)

    #fixing data for testing purposes
    # As = np.zeros((num_resources, 2, 2), dtype=np.float_)
    # As[0, :, :] = np.transpose([[0.87, 0.13], [0.04, 0.96]])
    # As[1, :, :] = np.transpose([[0.84, 0.16], [0.01, 0.99]])
    #
    # given_know[0] = [0.01, 0.04, 0.01, 0.14]
    # given_know[1] = [0.99, 0.96, 0.99, 0.86]
    #
    # given_notknow[0] = [0.84, 0.85, 0.97, 0.93]
    # given_notknow[1] = [0.16, 0.15, 0.03, 0.07]
    #
    # pi_0 = np.array([[0.99], [0.01]])
    #
    # emissions[0, :, :] = np.transpose([[0.84, 0.01], [0.16, 0.99]])
    # emissions[1, :, :] = np.transpose([[0.85, 0.04], [0.15, 0.96]])
    # emissions[2, :, :] = np.transpose([[0.97, 0.01], [0.03, 0.99]])
    # emissions[3, :, :] = np.transpose([[0.93, 0.14], [0.07, 0.86]])

    modelstruct = {}

    modelstruct['prior'] = pi_0[1][0]
    modelstruct['learns'] = As[:, 1, 0]
    modelstruct['forgets'] = As[:, 0, 1]
    modelstruct['guesses'] = given_notknow[1, :]
    modelstruct['slips'] = given_know[0, :]

    modelstruct['As'] = As
    modelstruct['emissions'] = emissions
    modelstruct['pi_0'] = pi_0

    return(modelstruct)


#########################################
# synthetic_data_helper.py              #
# synthetic_data_helper                 #
#                                       #
# @author Anirudhan Badrinath           #
# Last edited: 26 February 2020         #
#########################################

import numpy as np
from numpy.random import uniform
from time import time
from pyBKT.generate import synthetic_data_helper

def synthetic_data(model, lengths, resources = None):
    start = time()
    num_resources = len(model["learns"])
    bigT = sum(lengths)

    # assert model['learns'].shape == (1, ), "synthetic data generation does not support multilearn"
    # assert model['guesses'].shape == (1, ), "synthetic data generation does not support multiguess/slip"
    # assert model['slips'].shape == (1, ), "synthetic data generation does not support multiguess/slip"

    if resources is None:
        resources = np.random.randint(1, high = num_resources+1, size = bigT)
    if "As" not in model:
        model["As"] = np.array([[1-model["learns"], model["forgets"]], [model["learns"], 1-model["forgets"]]])

    if "pi_0" not in model:
        model["pi_0"] = np.array([[1-model["prior"]], [model["prior"]]])

    starts = np.cumsum(lengths)
    starts = (starts - lengths + 1)[: len(starts)]
    syn_data = synthetic_data_helper.create_synthetic_data(model, starts, lengths, resources)
    d = syn_data['data']
    syn_data["data"] = d + 1 

    # syn_data["data"][:, resources != 1] = 0 #no data emitted unless resource == 1

    datastruct = {}
    datastruct["stateseqs"] = syn_data["stateseqs"]
    datastruct["data"] = syn_data["data"]
    datastruct["starts"] = starts
    datastruct["lengths"] = lengths
    datastruct["resources"] = resources

    return datastruct

def create_synthetic_data(model, starts, lengths, resources):
    """ Randomly models synthetic through the preexisting MODEL, given the
        STARTS and RESOURCES. """
    learns, forgets, guesses, slips = \
            model['learns'], model['forgets'], model['guesses'], model['slips']
    inverted_guess = 1 - guesses
    num_res, num_subparts, num_seqs, num_guesses = \
            len(learns), len(slips), len(starts), len(guesses)
    use_ne = len(guesses) >= 1000
    initial_dist = np.array([1 - model['prior'], model['prior']])

    as_matrix = np.empty((2, 2 * num_res))
    interleave(as_matrix[0], 1 - learns, forgets)
    interleave(as_matrix[1], learns, 1 - forgets)

    req_length = lengths[:num_seqs]
    big_t = int(sum(req_length))

    all_stateseqs, all_data, result = \
            np.empty((1, big_t), dtype = np.int32), np.empty((num_subparts, big_t), dtype = np.int32), {}
    all_data[0][0] = 0
    for seq_index in range(num_seqs):
        seq_start, T = starts[seq_index] - 1, lengths[seq_index]
        nextstate_dist = initial_dist
        big_rand = uniform(size = num_guesses * T)
        other_rand = uniform(size = T)
        loop(big_rand, other_rand, T, num_guesses, nextstate_dist, seq_start, 
             slips, inverted_guess, all_data, all_stateseqs, as_matrix, resources)
    result['stateseqs'], result['data'] = all_stateseqs, all_data
    return result

def loop(big_rand, other_rand, T, num_guesses, nextstate_dist, seq_start, 
             slips, inverted_guess, all_data, all_stateseqs, as_matrix, resources):
  for t in range(T):
    k = seq_start + t
    r = resources[k]
    all_stateseqs[0][k] = a_sq = nextstate_dist[0] < other_rand[t]
    m = t * num_guesses
    rand_arr = big_rand[m: m + num_guesses]
    all_data[:, k] = rand_arr > (slips if a_sq else inverted_guess)
    get = r * 2 + a_sq - 2
    nextstate_dist = as_matrix[:, get]

def interleave(m, v1, v2):
    m[0::2], m[1::2] = v1, v2




import sys

def print_dot(num, tot, newline_every=None):
    if newline_every == None: newline_every = 25
    sys.stdout.write('.')

    if(num+1 % newline_every == 0 or num == tot-1):
        sys.stdout.write(' [%d/%d]\n' % (num+1, tot))
    sys.stdout.flush()


import numpy as np
import re
import sklearn.metrics as sk

def error_check(flat_true_values, pred_values):
    if len(flat_true_values) != len(pred_values):
        raise ValueError("preds and true values need to have same shape")

def accuracy(flat_true_values, pred_values):
    error_check(flat_true_values, pred_values)
    if len(flat_true_values) == 0:
        return np.nan
    correct = 0
    for i in range(len(pred_values)):
        if pred_values[i] >= 0.5 and flat_true_values[i] == 1:
            correct += 1
        if pred_values[i] < 0.5 and flat_true_values[i] == 0:
            correct += 1
    return correct/len([x for x in flat_true_values if (x == 0 or x == 1)])

def auc(flat_true_values, pred_values):
    error_check(flat_true_values, pred_values)
    # multiprior handling, remove phantom nondata
    if len(flat_true_values) == 0:
        return np.nan
    i = 0
    while i < len(flat_true_values):
        if (flat_true_values[i] != 1 and flat_true_values[i] != 0) or (pred_values[i] < 0 or pred_values[i] > 1):
            flat_true_values = np.delete(flat_true_values, i)
            pred_values = np.delete(pred_values, i)
            i -= 1
        i += 1
    if len(set(flat_true_values)) == 1:
        return np.nan

    auc = sk.roc_auc_score(flat_true_values, pred_values)
    return auc

def rmse(flat_true_values, pred_values):
    # represent correct as 1, incorrect as 0 for RMSE calculation
    if len(flat_true_values) == 0:
        return np.nan
    error_check(flat_true_values, pred_values)
    rmse, c = 0, 0
    for i in range(len(flat_true_values)):
        if flat_true_values[i] != -1:
            rmse += ((flat_true_values[i] - pred_values[i]) ** 2)
            c += 1
    rmse /= c
    rmse = rmse ** 0.5
    return rmse

def fetch_supported_metrics():
    supported_metrics = {}
    dummy_x, dummy_y = [0, 1] * 5, [1, 0] * 5
    for metric_locs in sk._regression, sk._classification:
        potential_metrics = {i: getattr(metric_locs, i) for i in dir(metric_locs) if re.search('_loss$|_score$|_error$', i)}
        for metric in potential_metrics:
            try:
                potential_metrics[metric](dummy_x, dummy_y)
                supported_metrics[metric] = potential_metrics[metric]
            except TypeError:
                pass
    return supported_metrics


SUPPORTED_METRICS = {'accuracy': accuracy, 'auc': auc, 'rmse': rmse}
SUPPORTED_METRICS.update(fetch_supported_metrics())



import sys, os
sys.path.append('..')
import numpy as np
from pyBKT.generate import synthetic_data, random_model_uni
from pyBKT.fit import EM_fit
from copy import deepcopy

#parameters classes
num_gs = 1 #number of guess/slip classes
num_learns = 1 #number of learning rates

num_fit_initializations = 20

#true params used for synthetic data generation
p_T = 0.30
p_F = 0.00
p_G = 0.10
p_S = 0.03
p_L0 = 0.10

#generate synthetic model and data.
truemodel = {}

truemodel["As"] =  np.zeros((num_learns,2,2), dtype=np.float_)
for i in range(num_learns):
    truemodel["As"][i] = np.transpose([[1-p_T, p_T], [p_F, 1-p_F]])

truemodel["learns"] = truemodel["As"][:,1, 0,]
truemodel["forgets"] = truemodel["As"][:,0, 1]

truemodel["pi_0"] = np.array([[1-p_L0], [p_L0]])
truemodel["prior"] = truemodel["pi_0"][1][0]

truemodel["guesses"] = np.full(num_gs, p_G, dtype=np.float_)
truemodel["slips"] = np.full(num_gs, p_S, dtype=np.float_)
#can optionally set learn class sequence - set randomly by synthetic_data if not included
#truemodel["resources"] = np.random.randint(1, high = num_resources, size = sum(observation_sequence_lengths))

#data!
print("generating data...")
observation_sequence_lengths = np.full(500, 100, dtype=np.int) #specifies 500 students with 100 observations for synthetic data
data = synthetic_data.synthetic_data(truemodel, observation_sequence_lengths)

#fit models, starting with random initializations
print('fitting! each dot is a new EM initialization')

num_fit_initializations = 5
best_likelihood = float("-inf")

for i in range(num_fit_initializations):
	fitmodel = random_model_uni.random_model_uni(num_learns, num_gs) # include this line to randomly set initial param values
	(fitmodel, log_likelihoods) = EM_fit.EM_fit(fitmodel, data)
	if(log_likelihoods[-1] > best_likelihood):
		best_likelihood = log_likelihoods[-1]
		best_model = fitmodel

# compare the fit model to the true model

print('')
print('\ttruth\tlearned')
print('prior\t%.4f\t%.4f' % (truemodel['prior'], best_model["pi_0"][1][0]))
for r in range(num_learns):
    print('learn%d\t%.4f\t%.4f' % (r+1, truemodel['As'][r, 1, 0].squeeze(), best_model['As'][r, 1, 0].squeeze()))
for r in range(num_learns):
    print('forget%d\t%.4f\t%.4f' % (r+1, truemodel['As'][r, 0, 1].squeeze(), best_model['As'][r, 0, 1].squeeze()))

for s in range(num_gs):
    print('guess%d\t%.4f\t%.4f' % (s+1, truemodel['guesses'][s], best_model['guesses'][s]))
for s in range(num_gs):
    print('slip%d\t%.4f\t%.4f' % (s+1, truemodel['slips'][s], best_model['slips'][s]))



import numpy as np
from pyBKT.generate import synthetic_data
from pyBKT.fit import predict_onestep

#parameters
num_subparts = 1
num_resources = 2
num_fit_initializations = 10
observation_sequence_lengths = np.full(500, 100, dtype=np.int)

#generate synthetic model and data.
#model is really easy.
truemodel = {}

truemodel["As"] =  np.zeros((2, 2, num_resources), dtype=np.float_)
for i in range(num_resources):
    truemodel["As"][i, :, :] = np.transpose([[0.7, 0.3], [0.01, 0.99]])
truemodel["learns"] = truemodel["As"][:, 1, 0]
truemodel["forgets"] = truemodel["As"][:, 0, 1]

truemodel["pi_0"] = np.array([[0.9], [0.1]])
truemodel["prior"] = truemodel["pi_0"][1][0]

truemodel["guesses"] = np.full(num_subparts, 0.1, dtype=np.float_)
truemodel["slips"] = np.full(num_subparts, 0.03, dtype=np.float_)

#data!
print("generating data...")
data = synthetic_data.synthetic_data(truemodel, observation_sequence_lengths)

(correct_predictions, state_predictions) = predict_onestep.run(truemodel, data)

print(correct_predictions, state_predictions)

print("finishing...")



import numpy as np

def run(model, trans_softcounts, emission_softcounts, init_softcounts):

    z = np.sum(trans_softcounts, axis=1) == 0
    for i in range(len(z)):
        for j in range(len(z[0])):
            if z[i,j]:
                trans_softcounts[i, 0, j] = 0
                trans_softcounts[i, 1, j] = 1
    # trans_softcounts[np.sum(trans_softcounts, axis=1) == 0,:] = 1
    emission_softcounts[np.sum(emission_softcounts, axis=2) == 0,:] = 1
    assert (trans_softcounts.shape[1] == 2)
    assert (trans_softcounts.shape[2] == 2)

    temp = np.sum(trans_softcounts, axis=1)

    #model['As'] = trans_softcounts / np.sum(trans_softcounts, axis=1)
    #model['As'] = np.divide(trans_softcounts, np.sum(trans_softcounts, axis=1))
    model['As'][:model['As'].shape[0]] = (trans_softcounts / np.sum(trans_softcounts, axis=1)[:model['As'].shape[0], None])

    model['learns'] = model['As'][:, 1, 0]
    model['forgets'] = model['As'][:, 0, 1]

    temp = np.sum(emission_softcounts, axis=2)

    #model['emissions'] = emission_softcounts / np.sum(emission_softcounts, axis=1)
    model['emissions'] = emission_softcounts / temp[:, :, None]
    #the expand dims is very weird.
    #model['guesses'] = np.expand_dims(model['emissions'][:, 0, 1].squeeze(), axis=0)
    #model['slips'] = np.expand_dims(model['emissions'][:, 1, 0].squeeze(), axis=0)
    model['guesses'] = model['emissions'][:, 0, 1]
    model['slips'] = model['emissions'][:, 1, 0]

    temp = np.sum(init_softcounts[:])
    model['pi_0'] = init_softcounts[:] / np.sum(init_softcounts[:])
    model['prior'] = model['pi_0'][1][0]

    return(model)



#########################################
# predict_onestep.py                    #
# predict_onestep                       #
#                                       #
# @author Anirudhan Badrinath           #
# @author Christian Garay               #
# Last edited: 20 March 2020            #
#########################################

import numpy as np
from pyBKT.fit.EM_fit import run as E_step_run

# correct_emission_predictions is a  num_subparts x T array, where element
# (i,t) is predicted probability that answer to subpart i at time t+1 is correct
def run(model, data):

    num_subparts = data["data"].shape[0]  # mmm the first dimension of data represents each subpart?? interesting.
    num_resources = len(model["learns"])

    trans_softcounts = np.zeros((num_resources, 2, 2))
    emission_softcounts = np.zeros((num_subparts, 2, 2))
    init_softcounts = np.zeros((2, 1))

    result = {}
    result['all_trans_softcounts'] = trans_softcounts
    result['all_emission_softcounts'] = emission_softcounts
    result['all_initial_softcounts'] = init_softcounts

    result = E_step_run(data, model, result['all_trans_softcounts'], result['all_emission_softcounts'], result['all_initial_softcounts'], 1)
    for j in range(num_resources):
        result['all_trans_softcounts'][j] = result['all_trans_softcounts'][j].transpose()
    for j in range(num_subparts):
        result['all_emission_softcounts'][j] = result['all_emission_softcounts'][j].transpose()
    state_predictions = predict_onestep_states(data, model, result['alpha_out'])
    # multiguess solution, should work
    correct_emission_predictions = np.expand_dims(model["guesses"], axis = 1) @ np.expand_dims(state_predictions[0,:], axis = 0) + np.expand_dims(1-model["slips"], axis = 1) @ np.expand_dims(state_predictions[1,:], axis = 0)
    #correct_emission_predictions = model['guesses'] * np.asarray([state_predictions[0,:]]).T + (1 - model['slips']) * np.asarray([state_predictions[1,:]]).T
    flattened_predictions = np.take_along_axis(correct_emission_predictions, (data['data'] != 0).argmax(axis = 0)[:, None].T, axis = 0)
    return (flattened_predictions.ravel(), state_predictions)

def predict_onestep_states(data, model, forward_messages):
    alldata, allresources, starts, lengths, learns, forgets, guesses, slips, prior = \
            data["data"], data["resources"], data["starts"], data["lengths"], \
            model["learns"], model["forgets"], model["guesses"], model["slips"], \
            model["prior"]
    bigT, num_subparts, num_sequences, num_resources = \
            len(alldata[0]), len(alldata), len(starts), len(learns)

    initial_distn = np.array([1 - prior, prior])

    As = np.empty((2, 2 * num_resources))
    interleave(As[0], 1 - learns, forgets)
    interleave(As[1], learns, 1 - forgets)

    fd_temp = np.empty((2 * bigT, ))
    for i in range(2):
        for j in range(bigT):
            fd_temp[i * bigT + j] = forward_messages[i][j]

    # outputs
    all_predictions = np.empty((2 * bigT, ))
    for sequence_index in range(num_sequences):
        sequence_start = starts[sequence_index] - 1
        T = lengths[sequence_index]
        forward_messages = fd_temp[2 * sequence_start: 2 * (sequence_start + T)].reshape((2, T), order = 'F')
        predictions = all_predictions[2 * sequence_start: 2 * (sequence_start + T)].reshape((2, T), order = 'F')

        predictions[:, 0] = initial_distn
        for t in range(T - 1):
            resources_temp = allresources[sequence_start+t]
            k = 2 * (resources_temp - 1)
            predictions[:, t + 1] = As[0: 2, k: k + 2].dot(forward_messages[:, t])

    return all_predictions.reshape((2, bigT), order = 'F')

def interleave(m, v1, v2):
        m[0::2], m[1::2] = v1, v2



#########################################
# data_helper.py                        #
# data_helper                           #
#                                       #
# @author Frederic Wang                 #
# Last edited: 27 March 2021            #
#########################################

import numpy as np
from pyBKT.generate import synthetic_data, random_model_uni
from pyBKT.fit import EM_fit, predict_onestep
from copy import deepcopy

# returns data only for the indices given based on starts array
def fix_data(data, indices):
    training_data = {}
    prev_starts = data["starts"][indices]
    lengths = data["lengths"][indices]
    total_length = np.sum(lengths)
    d = np.zeros((len(data["data"]), total_length), dtype=np.int32)
    resources = np.ones(total_length, dtype=np.int64)
    
    if "resource_names" in data:
        training_data["resource_names"] = data["resource_names"]
    if "gs_names" in data:
        training_data["gs_names"] = data["gs_names"]
    
    starts = np.zeros(len(prev_starts), dtype=np.int64)
    current_index = 1
    for i in range(len(prev_starts)):
        starts[i] = current_index

        d[:,starts[i]-1:starts[i]+lengths[i]-1] = data["data"][:,prev_starts[i]-1:prev_starts[i]+lengths[i]-1]
        resources[starts[i]-1:starts[i]+lengths[i]-1] = data["resources"][prev_starts[i]-1:prev_starts[i]+lengths[i]-1]
        current_index += lengths[i]
    
    training_data["starts"] = starts
    training_data["lengths"] = lengths
    training_data["data"] = d
    training_data["resources"]=resources
    return (training_data)

def fix_data_specified(data, label, count):
    training_data = {}
    testing_data = {}
    
    train_starts = np.zeros(len(data["starts"]), dtype = 'int')
    train_lengths = np.zeros(len(data["lengths"]), dtype = 'int')
    train_resources = np.zeros(len(data["resources"]) - count, dtype = 'int')
    train_data = np.zeros((len(data["data"]), len(data["resources"]) - count), dtype = 'int')
    
    test_starts = np.zeros(len(data["starts"]), dtype = 'int')
    test_lengths = np.zeros(len(data["lengths"]), dtype = 'int')
    test_resources = np.zeros(count, dtype = 'int')
    test_data = np.zeros((len(data["data"]), count), dtype = 'int')
     
    train_idx = 1
    test_idx = 1
    for i in range(len(data["starts"])):
        train_starts[i] = train_idx
        test_starts[i] = test_idx
        for j in range(data["lengths"][i]):
            current_idx = data["starts"][i] + j - 1
            if data["folds"][current_idx] != label:
                save_idx = train_starts[i]+train_lengths[i]-1
                train_lengths[i] += 1
                train_data[:,save_idx] = data["data"][:,current_idx]
                train_resources[save_idx] = data["resources"][current_idx]
            else:
                save_idx = test_starts[i]+test_lengths[i]-1
                test_lengths[i] += 1
                test_data[:,save_idx] = data["data"][:,current_idx]
                test_resources[save_idx] = data["resources"][current_idx]
        train_idx += train_lengths[i]
        test_idx += test_lengths[i]
        
    real_train = np.nonzero(train_lengths)
    train_lengths = train_lengths[real_train]
    train_starts = train_starts[real_train]

    real_test = np.nonzero(test_lengths)
    test_lengths = test_lengths[real_test]
    test_starts = test_starts[real_test]
    
    training_data["starts"] = train_starts
    training_data["lengths"] = train_lengths
    training_data["data"] = train_data
    training_data["resources"] = train_resources
    
    if "resource_names" in data:
        training_data["resource_names"] = data["resource_names"]
        testing_data["resource_names"] = data["resource_names"]
    if "gs_names" in data:
        training_data["gs_names"] = data["gs_names"]
        testing_data["gs_names"] = data["gs_names"]
    
    testing_data["starts"] = test_starts
    testing_data["lengths"] = test_lengths
    testing_data["data"] = test_data
    testing_data["resources"] = test_resources
    
    
    
    return training_data, testing_data

def crossvalidate(model, data, skill, folds, metric, seed, use_folds=False):

    num_learns = len(data["resource_names"]) if "resource_names" in data else 1
    num_gs = len(data["gs_names"]) if "gs_names" in data else num_gs

    # create random permutation to act as indices for folds for crossvalidation
    shuffle = np.random.RandomState(seed=seed).permutation(len(data["starts"]))
    all_true, all_pred = [], []
    metrics = np.zeros((len(metric), ))

    
    if use_folds: # predetermined folds
        all_labels, all_counts = np.unique(data["folds"], return_counts=True)
        all_folds = dict(zip(all_labels, all_counts))
        folds = len(all_folds)
        
        for label, count in all_folds.items():
            training_data, test_data = fix_data_specified(data, label, count)
            model.fit_model[skill] = model._fit(training_data, skill, model.forgets)
            metrics += model._evaluate({skill: test_data}, metric)
            
    else: # crossvalidation on students which are identified by the starts array
        split_size = len(data["starts"]) // folds
        for iteration in range(folds):
            model.fit_model = {}
            # create training/test data based on random permutation from earlier
            train = np.concatenate((shuffle[0: iteration * split_size], shuffle[(iteration + 1) * split_size:
                                                                                len(data["starts"])]))
            training_data = fix_data(data, train)
            model.fit_model[skill] = model._fit(training_data, skill, model.forgets)

            test = shuffle[iteration*split_size:(iteration+1)*split_size]
            test_data = fix_data(data, test)
            metrics += model._evaluate({skill: test_data}, metric)

    return metrics / folds



#########################################
# Model.py                              #
# Model                                 #
#                                       #
# @author Anirudhan Badrinath           #
# Last edited: 07 April 2021            #
#########################################

import numpy as np
import numbers
import os
import pandas as pd
import random
import pickle
import urllib.request as urllib2
from pyBKT.generate import synthetic_data, random_model_uni
from pyBKT.fit import EM_fit, predict_onestep
from pyBKT.util import crossvalidate, data_helper, check_data, metrics

pd.options.display.float_format = '{:,.5f}'.format

class Model:
    MODELS_BKT = ['multilearn', 'multiprior', 'multipair', 'multigs']
    MODEL_ARGS = ['parallel', 'num_fits', 'seed', 'defaults'] + MODELS_BKT
    FIT_ARGS = ['skills', 'num_fits', 'defaults',
                            'parallel', 'forgets', 'preload'] + MODELS_BKT
    CV_ARGS = FIT_ARGS + ['folds', 'seed']
    DEFAULTS = {'num_fits': 5,
                'defaults': None,
                'parallel': True,
                'skills': '.*',
                'seed': random.randint(0, 1e8),
                'folds': 5,
                'forgets': False,
                'model_type': [False] * len(MODELS_BKT)}
    DEFAULTS_BKT = {'order_id': 'order_id',
                    'skill_name': 'skill_name',
                    'correct': 'correct',
                    'user_id': 'user_id',
                    'multilearn': 'template_id',
                    'multiprior': 'correct',
                    'multipair': 'problem_id',
                    'multigs': 'template_id',
                    'folds': 'template_id'}
    INITIALIZABLE_PARAMS = ['prior', 'learns', 'guesses', 'slips', 'forgets']

    def __init__(self, **kwargs):
        """
        Constructs a BKT Model. Takes arguments parallel, num_fits, seed, defaults,
        and any model variant(s) that may be used. Note that all of these can be modified
        during fit/crossvalidation time.

        >>> model = Model(seed = 42)
        >>> model
        Model(parallel=True, num_fits=5, seed=42, defaults=None)

        """
        self.fit_model = None
        self.manual_param_init = False
        self._check_args(Model.MODEL_ARGS, kwargs)
        self.keep = {}
        self._update_param(['parallel', 'num_fits', 'seed', 'defaults'], kwargs, keep = True)
        self._update_param('model_type', self._update_defaults(kwargs), keep = True)

    def fit(self, data_path = None, data = None, **kwargs):
        """
        Fits a BKT model given model and data information. Takes arguments skills,
        number of initialization fits, default column names (i.e. correct, skill_name),
        parallelization, and model types. Resets model state if uninitialized.

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.evaluate(data_path = 'as.csv', metric = 'auc')
        0.6128265543747811

        """
        if not self.manual_param_init:
            self.fit_model = {}
        self.partial_fit(data_path = data_path, data = data, **kwargs)

    def partial_fit(self, data_path = None, data = None, **kwargs):
        """
        Partially fits a BKT model given model and data information. Takes arguments skills,
        number of initialization fits, default column names (i.e. correct, skill_name),
        parallelization, and model types. Behaviour is to ignore if the model is changed
        between partial fits since parameters are copied but data is reprocessed. Note
        that model type need not be specified when using partial fit after the first partial
        fit.

        >>> model = Model(seed = 42)
        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.evaluate(data_path = 'as.csv', metric = 'auc')
        0.6579800168987382

        """
        self.manual_param_init = True
        self._check_data(data_path, data)
        self._check_args(Model.FIT_ARGS, kwargs)
        self._update_param(['skills', 'num_fits', 'defaults', 
                            'parallel', 'forgets'], kwargs)
        if self.fit_model is None or self.fit_model == {}:
            self.fit_model = {}
            self._update_param('model_type', self._update_defaults(kwargs))
        all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type)
        self._update_param(['skills'], {'skills': list(all_data.keys())})
        for skill in all_data:
            self.fit_model[skill] = self._fit(all_data[skill], skill, self.forgets, 
                                              preload = kwargs['preload'] if 'preload' in kwargs else False)
        self.manual_param_init = False

    def predict(self, data_path = None, data = None):
        """
        Predicts using the trained BKT model and test data information. Takes test data
        location or DataFrame as arguments. Returns a dictionary mapping skills to predicted
        values for those skills. Note that the predicted values are a tuple of
        (correct_predictions, state_predictions).

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> preds_df = model.predict(data_path = 'as.csv')
        >>> preds_df[preds_df['skill_name'] == 'Box and Whisker'][['user_id', 'correct', 'correct_predictions', 'state_predictions']]
              user_id  correct  correct_predictions  state_predictions
        0       64525        1              0.69205            0.28276
        1       64525        1              0.80226            0.10060
        2       70363        0              0.69205            0.28276
        3       70363        1              0.54989            0.51775
        4       70363        0              0.74196            0.20028
        ...       ...      ...                  ...                ...
        3952    96297        1              0.84413            0.03139
        3953    96297        1              0.84429            0.03113
        3954    96297        1              0.84432            0.03108
        3955    96298        1              0.69205            0.28276
        3956    96298        1              0.80226            0.10060

        [3957 rows x 4 columns]

        """
        self._check_data(data_path, data)
        if self.fit_model is None:
            raise ValueError("model has not been fitted yet")
        all_data, df = self._data_helper(data_path = data_path, data = data,
                             defaults = self.defaults, skills = self.skills,
                             model_type = self.model_type, gs_ref = self.fit_model,
                             resource_ref = self.fit_model,
                             return_df = True)
        # default best effort prediction of 0.5
        df['correct_predictions'] = 0.5
        df['state_predictions'] = 0.5
        for skill in all_data:
            correct_predictions, state_predictions = self._predict(self.fit_model[skill], all_data[skill])
            df.loc[all_data[skill]['index'], 'correct_predictions'] = correct_predictions
            df.loc[all_data[skill]['index'], 'state_predictions'] = state_predictions[0]
        return df

    def evaluate(self, data = None, data_path = None, metric = metrics.rmse):
        """
        Evaluates a BKT model given model and data information. Takes a metric and
        data location or DataFrame as arguments. Returns the value of the metric
        for the given trained model tested on the given data.

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.evaluate(data_path = 'as.csv', metric = 'auc')
        0.6128265543747811

        """
        self._check_data(data_path, data)
        if not isinstance(metric, (tuple, list)):
            metric = [metric]
        if self.fit_model is None:
            raise ValueError("model has not been fitted yet")
        else:
            for i in range(len(metric)):
                m = metric[i]
                if isinstance(m, str):
                    if not m in metrics.SUPPORTED_METRICS:
                        raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
                    metric[i] = metrics.SUPPORTED_METRICS[m] 
                elif not callable(m):
                    raise ValueError("metric must either be a string, function or list/tuple of strings and functions")

        all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type,
                                     gs_ref = self.fit_model, resource_ref = self.fit_model)
        results = self._evaluate(all_data, metric)
        return results[0] if len(results) == 1 else results

    def crossvalidate(self, data = None, data_path = None, metric = metrics.rmse, **kwargs):
        """
        Crossvalidates (trains and evaluates) the BKT model. Takes the data, metric, and any
        arguments that would be passed to the fit function (skills, number of initialization fits, 
        default column names, parallelization, and model types) as arguments.

        >>> model = Model(seed = 42)
        >>> model.crossvalidate(data_path = 'as.csv')
                                                            mean_error
        skill
        Circle Graph                                           0.45840
        Percent Of                                             0.38005
        Finding Percents                                       0.42757
        Equivalent Fractions                                   0.45754
        Proportion                                             0.45437
        ...                                                        ...
        Solving Systems of Linear Equations                    0.38976
        Simplifying Expressions positive exponents             0.46494
        Parts of a Polyomial, Terms, Coefficient, Monom...     0.33278
        Finding Slope From Equation                            0.30684
        Recognize Quadratic Pattern                            0.00000

        [110 rows x 1 columns]

        """
        metric_names = []
        if not isinstance(metric, (tuple, list)):
            metric = [metric]
        if not isinstance(data, pd.DataFrame) and not isinstance(data_path, str):
            raise ValueError("no data specified")
        else:
            for i in range(len(metric)):
                m = metric[i]
                if isinstance(m, str):
                    if not m in metrics.SUPPORTED_METRICS:
                        raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
                    metric[i] = metrics.SUPPORTED_METRICS[m]
                    metric_names.append(m)
                elif callable(m):
                    metric_names.append(m.__name__)
                else:
                    raise ValueError("metric must either be a string, function or list/tuple of strings and functions")

        self._check_args(Model.CV_ARGS, kwargs)
        self._update_param(['skills', 'num_fits', 'defaults', 
                            'parallel', 'forgets', 'seed', 'folds'], kwargs)
        self._update_param('model_type', self._update_defaults(kwargs))
        metric_vals = {}
        if not self.manual_param_init:
            self.fit_model = {}
        if isinstance(self.folds, str):
            self._update_defaults({'folds': self.folds})
        all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type, folds = isinstance(self.folds, str))
        self._update_param(['skills'], {'skills': list(all_data.keys())})
        for skill in all_data:
            metric_vals[skill] = self._crossvalidate(all_data[skill], skill, metric)
        self.manual_param_init = False
        df = pd.DataFrame(metric_vals.items())
        df.columns = ['skill', 'dummy']
        df[metric_names] = pd.DataFrame(df['dummy'].tolist(), index = df.index)
        return df.set_index('skill').drop(columns = 'dummy')

    @property
    def coef_(self):
        """
        Returns the learned or preset parameters in the BKT model. Errors if model
        has not been fit or initialized. Note that the parameters are unique for
        each trained skill.

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.coef_
        {'Box and Whisker': {'learns': array([0.17969027]), 'forgets': array([0.01269486]), 'guesses': array([0.26595481]), 'slips': array([0.14831746]), 'prior': 0.8268892896231745}}

        """
        if not self.fit_model:
            raise ValueError("model has not been trained or initialized")
        return {skill: {param: self.fit_model[skill][param] for param in Model.INITIALIZABLE_PARAMS
                        if param in self.fit_model[skill]}
                for skill in self.fit_model}

    @coef_.setter
    def coef_(self, values):
        """
        Sets or initializes parameters in the BKT model. Values must be organized
        by skill and the BKT parameters as follows: {skill_name: {'learns': ..., 'guesses': ...}.
        Note that all parameters except the prior must be NumPy arrays.

        >>> model = Model(seed = 42)
        >>> model.coef_ = {'Box and Whisker': {'prior': 0.5}}
        >>> model.coef_
        {'Box and Whisker': {'prior': 0.5}}
        >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
        >>> model.coef_
        {'Box and Whisker': {'prior': 0.8221172842316857, 'learns': array([0.17918678]), 'guesses': array([0.27305474]), 'slips': array([0.14679578]), 'forgets': array([0.01293728])}}

        """
        self.fit_model = {}
        for skill in values:
            if skill not in self.fit_model:
                self.fit_model[skill] = {}
            if not self._check_params(values[skill]):
                raise ValueError("error in length, type or non-existent parameter")
            for param in values[skill]:
                self.fit_model[skill][param] = values[skill][param]
        self.manual_param_init = True

    def params(self):
        """ 
        Returns a DataFrame containing fitted parameters for easy
        printing.

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', multilearn = True, forgets = True, skills = 'Box and Whisker')
        >>> model.params()
                                          value
        skill           param   class          
        Box and Whisker prior   default 0.67443
                        learns  30799   0.16737
                                30059   0.33788
                                30060   0.28723
                                63448   0.10231
                                63447   0.07025
                                63446   0.13453
                        guesses default 0.31793
                        slips   default 0.12543
                        forgets 30799   0.00000
                                30059   0.04908
                                30060   0.01721
                                63448   0.03895
                                63447   0.00000
                                63446   0.01058

        """
        coefs = self.coef_
        formatted_coefs = []
        for skill in coefs:
            for param in coefs[skill]:
                classes = self._format_param(skill, param, coefs[skill][param])
                for class_ in classes:
                   formatted_coefs.append((skill, param, str(class_), classes[class_]))
        df = pd.DataFrame(formatted_coefs)
        df.columns = ['skill', 'param', 'class', 'value']
        return df.set_index(['skill', 'param', 'class'])

    def save(self, loc):
        """
        Saves a model to disk. Uses Python pickles.

        >>> model = Model(seed = 42)
        >>> model.fit(data_path = 'as.csv', multilearn = True, forgets = True, skills = 'Box and Whisker')
        >>> model.save('model.pkl')
        """
        with open(loc, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def load(self, loc):
        """
        Loads model given by loc into the current model.

        >>> model = Model(seed = 42)
        >>> model.load('model.pkl')
        """
        with open(loc, 'rb') as handle:
            orig_model = pickle.load(handle)
        for attr in vars(orig_model):
            setattr(self, attr, getattr(orig_model, attr))


    def fetch_dataset(self, link, loc):
        """
        Fetches dataset from an online link. Must be accessible without password
        or other authentication. Saves to the given location.

        >>> model = Model()
        >>> model.fetch_dataset('https://raw.githubusercontent.com/CAHLR/pyBKT-examples/master/data/as.csv', '.')
        """
        file_data = urllib2.urlopen(link)
        name = link.split('/')[-1]
        with open(os.path.normpath(loc + '/' + name), 'wb') as f:
            f.write(file_data.read())

    def _data_helper(self, data_path, data, defaults, skills, model_type, gs_ref = None, resource_ref = None, return_df = False, folds = False):
        """ Processes data given defaults, skills, and the model type. """
        if isinstance(data_path, str):
            data_p = data_helper.convert_data(data_path, skills, defaults = defaults, model_type = model_type, 
                                              gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df, folds = folds)
        elif isinstance(data, pd.DataFrame):
            data_p = data_helper.convert_data(data, skills, defaults = defaults, model_type = model_type,
                                                gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df, folds = folds)
        if not return_df:
            for d in data_p.values():
                check_data.check_data(d)
        else:
            for d in data_p[0].values():
                check_data.check_data(d)
        return data_p

    def _fit(self, data, skill, forgets, preload = False):
        """ Helper function for fitting data. """
        num_learns = len(data["resource_names"])
        num_gs = len(data["gs_names"])
        self._check_manual_param_init(num_learns, num_gs, skill)
        num_fit_initializations = self.num_fits
        best_likelihood = float("-inf")
        best_model = None

        for i in range(num_fit_initializations):
            fitmodel = random_model_uni.random_model_uni(num_learns, num_gs)
            if forgets:
                fitmodel["forgets"] = np.random.uniform(size = fitmodel["forgets"].shape)
            if self.manual_param_init and skill in self.fit_model:
                for var in self.fit_model[skill]:
                    if var in fitmodel:
                        fitmodel[var] = self.fit_model[skill][var]
            if not preload:
                fitmodel, log_likelihoods = EM_fit.EM_fit(fitmodel, data, parallel = self.parallel)
                if log_likelihoods[-1] > best_likelihood:
                    best_likelihood = log_likelihoods[-1]
                    best_model = fitmodel
            else:
                best_model = fitmodel
        fit_model = best_model
        fit_model["learns"] = fit_model["As"][:, 1, 0]
        fit_model["forgets"] = fit_model["As"][:, 0, 1]
        fit_model["prior"] = fit_model["pi_0"][1][0]
        fit_model["resource_names"] = data["resource_names"]
        fit_model["gs_names"] = data["gs_names"]
        return fit_model
    
    def _predict(self, model, data):
        """ Helper function for predicting. """
        return predict_onestep.run(model, data)

    def _evaluate(self, all_data, metric):
        """ Helper function for evaluating. """
        per_skill = []
        true, pred = [], []
        for skill in all_data:
            correct_predictions, state_predictions = self._predict(self.fit_model[skill], all_data[skill])
            real_data = all_data[skill]['data']
            true = np.append(true, real_data.sum(axis = 0))
            pred = np.append(pred, correct_predictions)
        true = true - 1
        try:
            res = [m(true, pred) for m in metric]
        except ValueError:
            res = [m(true, pred.round(0)) for m in metric]
        return res

    def _crossvalidate(self, data, skill, metric):
        """ Helper function for crossvalidating. """
        if isinstance(self.folds, str):
            return crossvalidate.crossvalidate(self, data, skill, self.folds, metric, self.seed, True)
        else:
            return crossvalidate.crossvalidate(self, data, skill, self.folds, metric, self.seed)

    def _format_param(self, skill, param, value):
        """ Formats parameter for nice printing. """
        if isinstance(value, np.ndarray):
            ptype = 'resource_names' if (param == 'learns' or param == 'forgets') \
                                     else 'gs_names'
            names = [str(i) for i in self.fit_model[skill][ptype]]
            return dict(sorted(zip(names, value)))
        else:
            return {'default': value}

    def _update_param(self, params, args, keep = False):
        """ Updates parameters given kwargs. """
        if isinstance(args, dict):
            for param in params:
                if param not in args and (param not in self.keep or not self.keep[param]):
                    setattr(self, param, Model.DEFAULTS[param])
                elif param in args:
                    setattr(self, param, args[param])
                self.keep[param] = keep
        else:
            setattr(self, params, args)
            self.keep[params] = keep

    def _update_defaults(self, defaults):
        """ Update the default column names. """
        model_types = [False] * 4
        for d in defaults:
            if d in Model.MODELS_BKT:
                if isinstance(defaults[d], bool):
                    model_types[Model.MODELS_BKT.index(d)] = defaults[d]
                elif isinstance(defaults[d], str):
                    if self.defaults is None:
                        self.defaults = {}
                    self.defaults[d] = defaults[d]
                    model_types[Model.MODELS_BKT.index(d)] = True
                else:
                    raise ValueError("model type must either be boolean for automatic column inference" + \
                                     " or string specifying column")
            elif d in Model.DEFAULTS_BKT:
                if self.defaults is None:
                    self.defaults = {}
                self.defaults[d] = defaults[d]
        return model_types

    def _check_params(self, params):
        """ Checks if BKT parameters are valid. """
        valid = True
        for param in params:
            if param == 'prior':
                valid = valid and isinstance(params[param], numbers.Number)
            else:
                valid = valid and isinstance(params[param], np.ndarray) \
                                   and param in Model.INITIALIZABLE_PARAMS
        if 'learns' in params and 'forgets' in params:
            valid = valid and (len(params['learns']) == len(params['forgets']))
        if 'guesses' in params and 'slips' in params:
            valid = valid and (len(params['slips']) == len(params['guesses']))
        return valid

    def _check_manual_param_init(self, num_learns, num_gs, skill):
        if self.fit_model and skill in self.fit_model and 'learns' in self.fit_model[skill] \
                and len(self.fit_model[skill]['learns']) != num_learns:
            raise ValueError("invalid number of learns in initialization")
        if self.fit_model and skill in self.fit_model and 'guesses' in self.fit_model[skill] \
                and len(self.fit_model[skill]['guesses']) != num_gs:
            raise ValueError("invalid number of guess classes in initialization")
        if self.fit_model and skill in self.fit_model and 'slips' in self.fit_model[skill] \
                and len(self.fit_model[skill]['slips']) != num_gs:
            raise ValueError("invalid number of slip classes in initialization")

    def _check_args(self, expected_args, args):
        for arg in args:
            if arg not in expected_args:
                raise ValueError("provided arguments are not recognized. they must be one or more of: " + \
                        ", ".join(expected_args))

    def _check_data(self, data_path, data):
        if not isinstance(data_path, str) and not isinstance(data, pd.DataFrame):
            raise ValueError("no data specified")
        elif isinstance(data_path, str) and isinstance(data, pd.DataFrame):
            raise ValueError("cannot specify both data location and data")
        elif isinstance(data_path, str) and not os.path.exists(data_path):
            raise ValueError("data path is invalid or file not found")

    def __repr__(self):
        ret = 'Model('
        args = ['%s=%s' % (arg, str(getattr(self, arg))) for arg in Model.MODEL_ARGS if hasattr(self, arg)]
        ret += ', '.join(args) + ')'
        return ret



#########################################
# setup.py                              #
# Setup for PyBKT                       #
#                                       #
# @author Anirudhan Badrinath           #
# Last edited: 01 April 2021            #
#########################################

import os
from os.path import normpath as npath
import sys
from sysconfig import get_paths
from setuptools import setup, Extension
import platform
from distutils.command.build_ext import build_ext

sys.tracebacklimit = 0

class CustomBuildExtCommand(build_ext):
    """build_ext command for use when numpy headers are needed."""
    """ https://stackoverflow.com/questions/2379898/make-distutils-look-for-numpy-header-files-in-the-correct-place """
    def run(self):
        import numpy
        self.include_dirs.append(numpy.get_include())
        build_ext.run(self)

FILES = {'synthetic_data_helper.cpp': 'source-cpp/pyBKT/generate/',
         'predict_onestep_states.cpp': 'source-cpp/pyBKT/fit/', 
         'E_step.cpp': 'source-cpp/pyBKT/fit/'}

if platform.system() == 'Darwin':
    ALL_COMPILE_ARGS = ['-c', '-fPIC', '-w', '-O3', '-stdlib=libc++', '-Xpreprocessor', '-fopenmp']
    ALL_LINK_ARGS = ['-stdlib=libc++']
    ALL_LIBRARIES = ['pthread', 'dl', 'util', 'm', 'omp']
else:
    ALL_COMPILE_ARGS = ['-c', '-fPIC', '-w', '-fopenmp', '-O2']
    ALL_LINK_ARGS = ['-fopenmp']
    ALL_LIBRARIES = ['pthread', 'dl', 'util', 'm']
INCLUDE_DIRS = sys.path + ['source-cpp/pyBKT/Eigen/', get_paths()['include']]
LIBRARY_DIRS = [os.environ['LD_LIBRARY_PATH']] if 'LD_LIBRARY_PATH' in os.environ \
                                               else []
def clean():
    global LIBRARY_DIRS, ALL_LIBRARIES
    LIBRARY_DIRS = [i for i in LIBRARY_DIRS if i != ""]
    ALL_LIBRARIES = [i for i in ALL_LIBRARIES if i != ""]

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

LIBRARY_DIRS += [sys.exec_prefix + '/lib']
clean()
try:
    module1 = Extension('pyBKT/generate/synthetic_data_helper',
                        sources = [npath('source-cpp/pyBKT/generate/synthetic_data_helper.cpp')], 
                        include_dirs = INCLUDE_DIRS,
                        extra_compile_args = ALL_COMPILE_ARGS,
                        library_dirs = LIBRARY_DIRS, 
                        libraries = ALL_LIBRARIES, 
                        extra_link_args = ALL_LINK_ARGS)

    module2 = Extension('pyBKT/fit/E_step', 
                        sources = [npath('source-cpp/pyBKT/fit/E_step.cpp')],
                        include_dirs = INCLUDE_DIRS,
                        extra_compile_args = ALL_COMPILE_ARGS,
                        library_dirs = LIBRARY_DIRS, 
                        libraries = ALL_LIBRARIES, 
                        extra_link_args = ALL_LINK_ARGS)

    module3 = Extension('pyBKT/fit/predict_onestep_states',
                        sources = [npath('source-cpp/pyBKT/fit/predict_onestep_states.cpp')],
                        include_dirs = INCLUDE_DIRS,
                        extra_compile_args = ALL_COMPILE_ARGS,
                        library_dirs = LIBRARY_DIRS, 
                        libraries = ALL_LIBRARIES, 
                        extra_link_args = ALL_LINK_ARGS)

    setup(
        name="pyBKT",
        version="1.3",
        author="Zachary Pardos, Anirudhan Badrinath, Matthew Jade Johnson, Christian Garay",
        author_email="zp@berkeley.edu, abadrinath@berkeley.edu, mattjj@csail.mit.edu, c.garay@berkeley.edu",
        license = 'MIT',
        description="PyBKT - Python Implentation of Bayesian Knowledge Tracing",
        url="https://github.com/CAHLR/pyBKT",
        download_url = 'https://github.com/CAHLR/pyBKT/archive/1.0.tar.gz',
        keywords = ['BKT', 'Bayesian Knowledge Tracing', 'Bayesian Network', 'Hidden Markov Model', 'Intelligent Tutoring Systems', 'Adaptive Learning'],
        classifiers=[
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        long_description = long_description,
        long_description_content_type='text/markdown',
        packages=['pyBKT', 'pyBKT.generate', 'pyBKT.fit', 'pyBKT.util', 'pyBKT.models'],
        package_dir = { 'pyBKT': npath('source-cpp/pyBKT'),
                        'pyBKT.generate': npath('source-cpp/pyBKT/generate'),
                        'pyBKT.fit': npath('source-cpp/pyBKT/fit'),
                        'pyBKT.util': npath('source-cpp/pyBKT/util'),
                        'pyBKT.models': npath('source-cpp/pyBKT/models')},
        install_requires = ["numpy", "sklearn", "pandas", "requests"],
        setup_requires = ["numpy"],
        cmdclass = {'build_ext': CustomBuildExtCommand},
        ext_modules = [module1, module2, module3]
    )
except:
# LEGACY PURE PYTHON VERSION:
    setup(
        name="pyBKT",
        version="1.3",
        author="Zachary Pardos, Anirudhan Badrinath, Matthew Jade Johnson, Christian Garay",
        author_email="zp@berkeley.edu, abadrinath@berkeley.edu, mattjj@csail.mit.edu, c.garay@berkeley.edu",
        license = 'MIT',
        description="PyBKT - Python Implentation of Bayesian Knowledge Tracing",
        url="https://github.com/CAHLR/pyBKT",
        download_url = 'https://github.com/CAHLR/pyBKT/archive/1.0.tar.gz',
        keywords = ['BKT', 'Bayesian Knowledge Tracing', 'Bayesian Network', 'Hidden Markov Model', 'Intelligent Tutoring Systems', 'Adaptive Learning'],
        classifiers=[
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        long_description = long_description,
        long_description_content_type='text/markdown',
        packages=['pyBKT', 'pyBKT.generate', 'pyBKT.fit', 'pyBKT.util', 'pyBKT.models'],
        package_dir = { 'pyBKT': npath('source-py/pyBKT'),
                        'pyBKT.generate': npath('source-py/pyBKT/generate'),
                        'pyBKT.fit': npath('source-py/pyBKT/fit'),
                        'pyBKT.util': npath('source-py/pyBKT/util'),
                        'pyBKT.models': npath('source-py/pyBKT/models')},
        install_requires = ["numpy", "sklearn", "pandas", "requests"],
    )



import sys
sys.path.append('../')
import numpy as np
from pyBKT.generate import synthetic_data, random_model_uni
from pyBKT.fit import EM_fit
from pyBKT.util.data_helper import assistments_data
from copy import deepcopy

skill = 'Pythagorean Theorem'

#parameters classes
num_gs = 1 #number of guess/slip classes
num_learns = 1 #number of learning rates

num_fit_initializations = 20

#alternatively, you can load REAL data - might take a minute (involves dataset download each time the function is called)
print('loading assistments %s data...' % skill)
data = assistments_data(skill)


#fit models, starting with random initializations
print('fitting! each dot is a new EM initialization')

num_fit_initializations = 5
best_likelihood = float("-inf")

for i in range(num_fit_initializations):
	fitmodel = random_model_uni.random_model_uni(num_learns, num_gs) # include this line to randomly set initial param values
	(fitmodel, log_likelihoods) = EM_fit.EM_fit(fitmodel, data)
	if(log_likelihoods[-1] > best_likelihood):
		best_likelihood = log_likelihoods[-1]
		best_model = fitmodel

# compare the fit model to the true model

print('')
print('\tlearned')
print('prior\t%.4f' % (best_model["pi_0"][1][0]))
for r in range(num_learns):
    print('learn%d\t%.4f' % (r+1, best_model['As'][r, 1, 0].squeeze()))
for r in range(num_learns):
    print('forget%d\t%.4f' % (r+1, best_model['As'][r, 0, 1].squeeze()))

for s in range(num_gs):
    print('guess%d\t%.4f' % (s+1, best_model['guesses'][s]))
for s in range(num_gs):
    print('slip%d\t%.4f' % (s+1, best_model['slips'][s]))



import numpy as np

def dirrnd(alphavec):
    dimsum = 1
    if(alphavec.ndim == 3 or alphavec.ndim == 2): dimsum = alphavec.ndim - 2
    a = np.random.gamma(alphavec, 1)
    temp = np.sum(a, axis=dimsum, keepdims=True)
    a = a / np.sum(a, axis=dimsum, keepdims=True)
    return(a)


def check_data(data):
    if data["data"].shape[1] != data["resources"].shape[0]: raise IndexError("data and resource sizes must match")
    if not all(data["starts"] + data["lengths"] - 1 <= data["data"].shape[1]): raise IndexError("data lengths don't match its shape")
